@@ -18,10 +18,11 @@
 from . import mparser, mesonlib, mlog
 from . import environment, dependencies
 
-import abc
-import os, copy, re
-import collections.abc
 from functools import wraps
+import abc
+import collections.abc
+import itertools
+import os, copy, re
 import typing as T
 
 TV_fw_var = T.Union[str, int, float, bool, list, dict, 'InterpreterObject', 'ObjectHolder']
@@ -59,16 +60,30 @@ class InterpreterObject:
 TV_InterpreterObject = T.TypeVar('TV_InterpreterObject')
 
 class ObjectHolder(T.Generic[TV_InterpreterObject]):
-    def __init__(self, obj: InterpreterObject, subproject: T.Optional[str] = None) -> None:
-        self.held_object = obj        # type: InterpreterObject
-        self.subproject = subproject  # type: str
+    def __init__(self, obj: TV_InterpreterObject, subproject: str = '') -> None:
+        self.held_object = obj
+        self.subproject = subproject
 
     def __repr__(self) -> str:
-        return '<Holder: {!r}>'.format(self.held_object)
+        return f'<Holder: {self.held_object!r}>'
 
 class MesonVersionString(str):
     pass
 
+class RangeHolder(InterpreterObject):
+    def __init__(self, start: int, stop: int, step: int) -> None:
+        super().__init__()
+        self.range = range(start, stop, step)
+
+    def __iter__(self) -> T.Iterator[int]:
+        return iter(self.range)
+
+    def __getitem__(self, key: int) -> int:
+        return self.range[key]
+
+    def __len__(self) -> int:
+        return len(self.range)
+
 # Decorators for method calls.
 
 def check_stringlist(a: T.Any, msg: str = 'Arguments must be strings.') -> None:
@@ -100,7 +115,7 @@ def _get_callee_args(wrapped_args: T.Sequence[T.Any], want_subproject: bool = Fa
             kwargs = None
             subproject = wrapped_args[1].subproject
         else:
-            raise AssertionError('Unknown args: {!r}'.format(wrapped_args))
+            raise AssertionError(f'Unknown args: {wrapped_args!r}')
     elif n == 3:
         # Methods on objects (*Holder, MesonMain, etc) have 3 args: self, args, kwargs
         node = s.current_node
@@ -133,7 +148,7 @@ def _get_callee_args(wrapped_args: T.Sequence[T.Any], want_subproject: bool = Fa
         if want_subproject:
             subproject = wrapped_args[2].subproject
     else:
-        raise AssertionError('Unknown args: {!r}'.format(wrapped_args))
+        raise AssertionError(f'Unknown args: {wrapped_args!r}')
     # Sometimes interpreter methods are called internally with None instead of
     # empty list/dict
     args = args if args is not None else []
@@ -173,7 +188,7 @@ def builtinMethodNoKwargs(f: TV_func) -> TV_func:
         method_name = wrapped_args[2]
         kwargs = wrapped_args[4]
         if kwargs:
-            mlog.warning('Method {!r} does not take keyword arguments.'.format(method_name),
+            mlog.warning(f'Method {method_name!r} does not take keyword arguments.',
                          'This will become a hard error in the future',
                          location=node)
         return f(*wrapped_args, **wrapped_kwargs)
@@ -223,11 +238,132 @@ class permittedKwargs:
             s, node, args, kwargs, _ = _get_callee_args(wrapped_args)
             for k in kwargs:
                 if k not in self.permitted:
-                    mlog.warning('''Passed invalid keyword argument "{}".'''.format(k), location=node)
+                    mlog.warning(f'''Passed invalid keyword argument "{k}".''', location=node)
                     mlog.warning('This will become a hard error in the future.')
             return f(*wrapped_args, **wrapped_kwargs)
         return T.cast(TV_func, wrapped)
 
+
+def typed_pos_args(name: str, *types: T.Union[T.Type, T.Tuple[T.Type, ...]],
+                   varargs: T.Optional[T.Union[T.Type, T.Tuple[T.Type]]] = None,
+                   optargs: T.Optional[T.List[T.Union[T.Type, T.Tuple[T.Type]]]] = None,
+                   min_varargs: int = 0, max_varargs: int = 0) -> T.Callable[..., T.Any]:
+    """Decorator that types type checking of positional arguments.
+
+    This supports two different models of optional aguments, the first is the
+    variadic argument model. Variadic arguments are a possibly bounded,
+    possibly unbounded number of arguments of the same type (unions are
+    supported). The second is the standard default value model, in this case
+    a number of optional arguments may be provided, but they are still
+    ordered, and they may have different types.
+
+    This function does not support mixing variadic and default arguments.
+
+    :name: The name of the decorated function (as displayed in error messages)
+    :varargs: They type(s) of any variadic arguments the function takes. If
+        None the function takes no variadic args
+    :min_varargs: the minimum number of variadic arguments taken
+    :max_varargs: the maximum number of variadic arguments taken. 0 means unlimited
+    :optargs: The types of any optional arguments parameters taken. If None
+        then no optional paramters are taken.
+
+    Some examples of usage blow:
+    >>> @typed_pos_args('mod.func', str, (str, int))
+    ... def func(self, state: ModuleState, args: T.Tuple[str, T.Union[str, int]], kwargs: T.Dict[str, T.Any]) -> T.Any:
+    ...     pass
+
+    >>> @typed_pos_args('method', str, varargs=str)
+    ... def method(self, node: BaseNode, args: T.Tuple[str, T.List[str]], kwargs: T.Dict[str, T.Any]) -> T.Any:
+    ...     pass
+
+    >>> @typed_pos_args('method', varargs=str, min_varargs=1)
+    ... def method(self, node: BaseNode, args: T.Tuple[T.List[str]], kwargs: T.Dict[str, T.Any]) -> T.Any:
+    ...     pass
+
+    >>> @typed_pos_args('method', str, optargs=[(str, int), str])
+    ... def method(self, node: BaseNode, args: T.Tuple[str, T.Optional[T.Union[str, int]], T.Optional[str]], kwargs: T.Dict[str, T.Any]) -> T.Any:
+    ...     pass
+
+    When should you chose `typed_pos_args('name', varargs=str,
+    min_varargs=1)` vs `typed_pos_args('name', str, varargs=str)`?
+
+    The answer has to do with the semantics of the function, if all of the
+    inputs are the same type (such as with `files()`) then the former is
+    correct, all of the arguments are string names of files. If the first
+    argument is something else the it should be separated.
+    """
+    def inner(f: TV_func) -> TV_func:
+
+        @wraps(f)
+        def wrapper(*wrapped_args: T.Any, **wrapped_kwargs: T.Any) -> T.Any:
+            args = _get_callee_args(wrapped_args)[2]
+
+            # These are implementation programming errors, end users should never see them.
+            assert isinstance(args, list), args
+            assert max_varargs >= 0, 'max_varags cannot be negative'
+            assert min_varargs >= 0, 'min_varags cannot be negative'
+            assert optargs is None or varargs is None, \
+                'varargs and optargs not supported together as this would be ambiguous'
+
+            num_args = len(args)
+            num_types = len(types)
+            a_types = types
+
+            if varargs:
+                min_args = num_types + min_varargs
+                max_args = num_types + max_varargs
+                if max_varargs == 0 and num_args < min_args:
+                    raise InvalidArguments(f'{name} takes at least {min_args} arguments, but got {num_args}.')
+                elif max_varargs != 0 and (num_args < min_args or num_args > max_args):
+                    raise InvalidArguments(f'{name} takes between {min_args} and {max_args} arguments, but got {num_args}.')
+            elif optargs:
+                if num_args < num_types:
+                    raise InvalidArguments(f'{name} takes at least {num_types} arguments, but got {num_args}.')
+                elif num_args > num_types + len(optargs):
+                    raise InvalidArguments(f'{name} takes at most {num_types + len(optargs)} arguments, but got {num_args}.')
+                # Add the number of positional arguments required
+                if num_args > num_types:
+                    diff = num_args - num_types
+                    a_types = tuple(list(types) + list(optargs[:diff]))
+            elif num_args != num_types:
+                raise InvalidArguments(f'{name} takes exactly {num_types} arguments, but got {num_args}.')
+
+            for i, (arg, type_) in enumerate(itertools.zip_longest(args, a_types, fillvalue=varargs), start=1):
+                if not isinstance(arg, type_):
+                    if isinstance(type_, tuple):
+                        shouldbe = 'one of: {}'.format(", ".join(f'"{t.__name__}"' for t in type_))
+                    else:
+                        shouldbe = f'"{type_.__name__}"'
+                    raise InvalidArguments(f'{name} argument {i} was of type "{type(arg).__name__}" but should have been {shouldbe}')
+
+            # Ensure that we're actually passing a tuple.
+            # Depending on what kind of function we're calling the length of
+            # wrapped_args can vary.
+            nargs = list(wrapped_args)
+            i = nargs.index(args)
+            if varargs:
+                # if we have varargs we need to split them into a separate
+                # tuple, as python's typing doesn't understand tuples with
+                # fixed elements and variadic elements, only one or the other.
+                # so in that case we need T.Tuple[int, str, float, T.Tuple[str, ...]]
+                pos = args[:len(types)]
+                var = list(args[len(types):])
+                pos.append(var)
+                nargs[i] = tuple(pos)
+            elif optargs:
+                if num_args < num_types + len(optargs):
+                    diff =  num_types + len(optargs) - num_args
+                    nargs[i] = tuple(list(args) + [None] * diff)
+                else:
+                    nargs[i] = args
+            else:
+                nargs[i] = tuple(args)
+            return f(*nargs, **wrapped_kwargs)
+
+        return T.cast(TV_func, wrapper)
+    return inner
+
+
 class FeatureCheckBase(metaclass=abc.ABCMeta):
     "Base class for feature version checks"
 
@@ -296,7 +432,7 @@ class FeatureCheckBase(metaclass=abc.ABCMeta):
         def wrapped(*wrapped_args: T.Any, **wrapped_kwargs: T.Any) -> T.Any:
             subproject = _get_callee_args(wrapped_args, want_subproject=True)[4]
             if subproject is None:
-                raise AssertionError('{!r}'.format(wrapped_args))
+                raise AssertionError(f'{wrapped_args!r}')
             self.use(subproject)
             return f(*wrapped_args, **wrapped_kwargs)
         return T.cast(TV_func, wrapped)
@@ -322,14 +458,14 @@ class FeatureNew(FeatureCheckBase):
 
     @staticmethod
     def get_warning_str_prefix(tv: str) -> str:
-        return 'Project specifies a minimum meson_version \'{}\' but uses features which were added in newer versions:'.format(tv)
+        return f'Project specifies a minimum meson_version \'{tv}\' but uses features which were added in newer versions:'
 
     def log_usage_warning(self, tv: str) -> None:
         args = [
-            'Project targeting', "'{}'".format(tv),
+            'Project targeting', f"'{tv}'",
             'but tried to use feature introduced in',
-            "'{}':".format(self.feature_version),
-            '{}.'.format(self.feature_name),
+            f"'{self.feature_version}':",
+            f'{self.feature_name}.',
         ]
         if self.extra_message:
             args.append(self.extra_message)
@@ -354,10 +490,10 @@ class FeatureDeprecated(FeatureCheckBase):
 
     def log_usage_warning(self, tv: str) -> None:
         args = [
-            'Project targeting', "'{}'".format(tv),
+            'Project targeting', f"'{tv}'",
             'but tried to use feature deprecated since',
-            "'{}':".format(self.feature_version),
-            '{}.'.format(self.feature_name),
+            f"'{self.feature_version}':",
+            f'{self.feature_name}.',
         ]
         if self.extra_message:
             args.append(self.extra_message)
@@ -383,7 +519,7 @@ class FeatureCheckKwargsBase(metaclass=abc.ABCMeta):
         def wrapped(*wrapped_args: T.Any, **wrapped_kwargs: T.Any) -> T.Any:
             kwargs, subproject = _get_callee_args(wrapped_args, want_subproject=True)[3:5]
             if subproject is None:
-                raise AssertionError('{!r}'.format(wrapped_args))
+                raise AssertionError(f'{wrapped_args!r}')
             for arg in self.kwargs:
                 if arg not in kwargs:
                     continue
@@ -586,6 +722,8 @@ class InterpreterBase:
             return self.evaluate_indexing(cur)
         elif isinstance(cur, mparser.TernaryNode):
             return self.evaluate_ternary(cur)
+        elif isinstance(cur, mparser.FormatStringNode):
+            return self.evaluate_fstring(cur)
         elif isinstance(cur, mparser.ContinueNode):
             raise ContinueRequest()
         elif isinstance(cur, mparser.BreakNode):
@@ -633,7 +771,7 @@ class InterpreterBase:
             if isinstance(result, Disabler):
                 return result
             if not(isinstance(result, bool)):
-                raise InvalidCode('If clause {!r} does not evaluate to true or false.'.format(result))
+                raise InvalidCode(f'If clause {result!r} does not evaluate to true or false.')
             if result:
                 prev_meson_version = mesonlib.project_meson_versions[self.subproject]
                 if self.tmp_meson_version:
@@ -804,11 +942,29 @@ The result of this is undefined and will become a hard error in a future Meson r
         else:
             return self.evaluate_statement(node.falseblock)
 
+    @FeatureNew('format strings', '0.58.0')
+    def evaluate_fstring(self, node: mparser.FormatStringNode) -> TYPE_var:
+        assert(isinstance(node, mparser.FormatStringNode))
+
+        def replace(match: T.Match[str]) -> str:
+            var = str(match.group(1))
+            try:
+                val = self.variables[var]
+                if not isinstance(val, (str, int, float, bool)):
+                    raise InvalidCode(f'Identifier "{var}" does not name a formattable variable ' +
+                        '(has to be an integer, a string, a floating point number or a boolean).')
+
+                return str(val)
+            except KeyError:
+                raise InvalidCode(f'Identifier "{var}" does not name a variable.')
+
+        return re.sub(r'@([_a-zA-Z][_0-9a-zA-Z]*)@', replace, node.value)
+
     def evaluate_foreach(self, node: mparser.ForeachClauseNode) -> None:
         assert(isinstance(node, mparser.ForeachClauseNode))
         items = self.evaluate_statement(node.items)
 
-        if isinstance(items, list):
+        if isinstance(items, (list, RangeHolder)):
             if len(node.varnames) != 1:
                 raise InvalidArguments('Foreach on array does not unpack')
             varname = node.varnames[0]
@@ -823,7 +979,7 @@ The result of this is undefined and will become a hard error in a future Meson r
         elif isinstance(items, dict):
             if len(node.varnames) != 2:
                 raise InvalidArguments('Foreach on dict unpacks key and value')
-            for key, value in items.items():
+            for key, value in sorted(items.items()):
                 self.set_variable(node.varnames[0], key)
                 self.set_variable(node.varnames[1], value)
                 try:
@@ -893,7 +1049,7 @@ The result of this is undefined and will become a hard error in a future Meson r
                 # TODO maybe find a better solution
                 return iobject[index]  # type: ignore
             except IndexError:
-                # We are already checking for the existance of __getitem__, so this should be save
+                # We are already checking for the existence of __getitem__, so this should be save
                 raise InterpreterException('Index %d out of bounds of array of size %d.' % (index, len(iobject)))  # type: ignore
 
     def function_call(self, node: mparser.FunctionNode) -> T.Optional[TYPE_var]:
@@ -945,7 +1101,7 @@ The result of this is undefined and will become a hard error in a future Meson r
                 return Disabler()
         if method_name == 'extract_objects':
             if not isinstance(obj, ObjectHolder):
-                raise InvalidArguments('Invalid operation "extract_objects" on variable "{}"'.format(object_name))
+                raise InvalidArguments(f'Invalid operation "extract_objects" on variable "{object_name}"')
             self.validate_extraction(obj.held_object)
         obj.current_node = node
         return obj.method_call(method_name, args, kwargs)
@@ -1039,7 +1195,7 @@ The result of this is undefined and will become a hard error in a future Meson r
             try:
                 return int(obj)
             except Exception:
-                raise InterpreterException('String {!r} cannot be converted to int'.format(obj))
+                raise InterpreterException(f'String {obj!r} cannot be converted to int')
         elif method_name == 'join':
             if len(posargs) != 1:
                 raise InterpreterException('Join() takes exactly one argument.')
@@ -1070,6 +1226,13 @@ The result of this is undefined and will become a hard error in a future Meson r
                     raise InterpreterException('substring() argument must be an int')
                 end = posargs[1]
             return obj[start:end]
+        elif method_name == 'replace':
+            FeatureNew.single_use('str.replace', '0.58.0', self.subproject)
+            if len(posargs) != 2:
+                raise InterpreterException('replace() takes exactly two arguments.')
+            if not isinstance(posargs[0], str) or not isinstance(posargs[1], str):
+                raise InterpreterException('replace() requires that both arguments be strings')
+            return obj.replace(posargs[0], posargs[1])
         raise InterpreterException('Unknown method "%s" for a string.' % method_name)
 
     def format_string(self, templ: str, args: T.List[TYPE_nvar]) -> str:
@@ -1084,7 +1247,7 @@ The result of this is undefined and will become a hard error in a future Meson r
         def arg_replace(match: T.Match[str]) -> str:
             idx = int(match.group(1))
             if idx >= len(arg_strings):
-                raise InterpreterException('Format placeholder @{}@ out of range.'.format(idx))
+                raise InterpreterException(f'Format placeholder @{idx}@ out of range.')
             return arg_strings[idx]
 
         return re.sub(r'@(\d+)@', arg_replace, templ)
@@ -1161,12 +1324,12 @@ The result of this is undefined and will become a hard error in a future Meson r
                     return self.evaluate_statement(fallback)
                 return fallback
 
-            raise InterpreterException('Key {!r} is not in the dictionary.'.format(key))
+            raise InterpreterException(f'Key {key!r} is not in the dictionary.')
 
         if method_name == 'keys':
             if len(posargs) != 0:
                 raise InterpreterException('keys() takes no arguments.')
-            return list(obj.keys())
+            return sorted(obj.keys())
 
         raise InterpreterException('Dictionaries do not have a method called "%s".' % method_name)
 
@@ -1204,7 +1367,7 @@ The result of this is undefined and will become a hard error in a future Meson r
             raise InterpreterException('Kwargs argument must not contain a "kwargs" entry. Points for thinking meta, though. :P')
         for k, v in to_expand.items():
             if k in kwargs:
-                raise InterpreterException('Entry "{}" defined both as a keyword argument and in a "kwarg" entry.'.format(k))
+                raise InterpreterException(f'Entry "{k}" defined both as a keyword argument and in a "kwarg" entry.')
             kwargs[k] = v
         return kwargs
 
